@@ -2,7 +2,7 @@
 #
 # COPYRIGHT:
 #
-# This software is Copyright (c) 1996-2013 Best Practical Solutions, LLC
+# This software is Copyright (c) 1996-2024 Best Practical Solutions, LLC
 #                                          <sales@bestpractical.com>
 #
 # (Except where explicitly superseded by other copyright notices)
@@ -69,6 +69,7 @@ use warnings;
 use 5.010;
 
 use base qw(DBIx::SearchBuilder RT::Base);
+$DBIx::SearchBuilder::PREFER_BIND = 1 unless defined $ENV{SB_PREFER_BIND};
 
 use RT::Base;
 use DBIx::SearchBuilder "1.40";
@@ -168,8 +169,15 @@ sub _OrderByCF {
         ENTRYAGGREGATOR => 'AND'
     );
 
-    return { %$row, ALIAS => $CFvs,  FIELD => 'SortOrder' },
-           { %$row, ALIAS => $ocfvs, FIELD => 'Content' };
+    return { %$row, ALIAS => $CFvs, FIELD => 'SortOrder' },
+        {
+            %$row,
+            ALIAS => $ocfvs,
+            FIELD => 'Content',
+            blessed $cf && $cf->IsNumeric
+            ? ( FUNCTION => RT->DatabaseHandle->CastAsDecimal('Content') )
+            : ()
+        };
 }
 
 sub OrderByCols {
@@ -266,6 +274,22 @@ sub RecordClass {
     $_[0]->_SingularClass
 }
 
+=head2 ReportClass
+
+Returns report class name of this collection. E.g. report class of RT::Tickets
+is RT::Report::Tickets
+
+=cut
+
+sub ReportClass {
+    my $self = shift;
+    my $class = ref($self) || $self;
+    if ( $class =~ s/(?<=^RT::)/Report::/ ) {
+        return $class;
+    }
+    return undef;
+}
+
 =head2 RegisterCustomFieldJoin
 
 Takes a pair of arguments, the first a class name and the second a callback
@@ -379,7 +403,7 @@ sub _CustomFieldJoin {
         $self->Limit(
             LEFTJOIN        => $ocfvalias,
             FIELD           => 'CustomField',
-            VALUE           => $cf->id,
+            VALUE           => $cf->Disabled ? 0 : $cf->id,
             ENTRYAGGREGATOR => 'AND'
         );
     }
@@ -435,6 +459,12 @@ sub _CustomFieldJoinByName {
         CASESENSITIVE   => 0,
         VALUE           => $cf,
     );
+    $self->Limit(
+        LEFTJOIN        => $CFs,
+        ENTRYAGGREGATOR => 'AND',
+        FIELD           => 'Disabled',
+        VALUE           => 0,
+    );
 
     my $ocfvalias = $self->Join(
         TYPE   => 'LEFT',
@@ -469,6 +499,7 @@ sub _LimitCustomField {
                  OPERATOR     => '=',
                  KEY          => undef,
                  PREPARSE     => 1,
+                 QUOTEVALUE   => 1,
                  @_ );
 
     my $op     = delete $args{OPERATOR};
@@ -479,6 +510,13 @@ sub _LimitCustomField {
     my $cfkey  = delete $args{KEY};
     if (blessed($cf) and $cf->id) {
         $cfkey ||= $cf->id;
+
+        # Make sure we can really see $cf
+        unless ( $cf->CurrentUserHasRight('SeeCustomField') ) {
+            my $obj = RT::CustomField->new( RT->SystemUser );
+            $obj->Load( $cf->id );
+            $cf = $obj;
+        }
     } elsif ($cf =~ /^\d+$/) {
         # Intentionally load as the system user, so we can build better
         # queries; this is necessary as we don't have a context object
@@ -493,16 +531,41 @@ sub _LimitCustomField {
             $cfkey ||= "$ltype-$cf";
         }
     } else {
-        $cfkey ||= "$ltype-$cf";
+        # Resolve CF by name for better queries, like the above block.
+        my $cfs = RT::CustomFields->new( RT->SystemUser );
+        $cfs->LimitToLookupType($ltype);
+        $cfs->Limit(
+            FIELD         => 'Name',
+            VALUE         => $cf,
+            CASESENSITIVE => 0,
+        );
+        if ( $cfs->Count == 1 ) {
+            $cf = $cfs->Next;
+            $cfkey ||= $cf->id;
+        }
+        else {
+            $cfkey ||= "$ltype-$cf";
+        }
     }
 
     $args{SUBCLAUSE} ||= "cf-$cfkey";
 
 
     my $fix_op = sub {
+        my %args = @_;
+
+        if (   $args{'FIELD'} eq 'Content'
+            && blessed $cf
+            && $cf->IsNumeric
+            && ( !$args{QUOTEVALUE} || Scalar::Util::looks_like_number($args{'VALUE'}) ) )
+        {
+            $args{QUOTEVALUE} = 0;
+            $args{FUNCTION} = RT->DatabaseHandle->CastAsDecimal( "$args{ALIAS}.$args{FIELD}" );
+            return %args;
+        }
+
         return @_ unless RT->Config->Get('DatabaseType') eq 'Oracle';
 
-        my %args = @_;
         return %args unless $args{'FIELD'} eq 'LargeContent';
 
         my $op = $args{'OPERATOR'};
@@ -518,10 +581,74 @@ sub _LimitCustomField {
         return %args;
     };
 
+    # Special Limit (we can exit early)
+    # IS NULL and IS NOT NULL checks
+    if ( $op =~ /^IS( NOT)?$/i ) {
+        my ($ocfvalias, $CFs) = $self->_CustomFieldJoin( $cfkey, $cf, $ltype );
+        $self->_OpenParen( $args{SUBCLAUSE} );
+        $self->Limit(
+            %args,
+            ALIAS    => $ocfvalias,
+            FIELD    => ($column || 'id'),
+            OPERATOR => $op,
+            VALUE    => $value,
+        );
+        # See below for an explanation of this limit
+        $self->Limit(
+            ALIAS      => $CFs,
+            FIELD      => 'Name',
+            OPERATOR   => 'IS NOT',
+            VALUE      => 'NULL',
+            ENTRYAGGREGATOR => 'AND',
+            SUBCLAUSE  => $args{SUBCLAUSE},
+        ) if $CFs;
+        $self->_CloseParen( $args{SUBCLAUSE} );
+        return;
+    }
+
     ########## Content pre-parsing if we know things about the CF
     if ( blessed($cf) and delete $args{PREPARSE} ) {
         my $type = $cf->Type;
-        if ( $type eq 'IPAddress' ) {
+
+        if ( !$args{QUOTEVALUE} ) {
+            my ( $class, $field );
+
+            # e.g. Users_3.Name
+            if ( $value =~ /^(\w+?)(?:_\d+)?\.(\w+)$/ ) {
+                my $table = $1;
+                $field = $2;
+                $class = $table =~ /main/i ? 'RT::Tickets' : "RT::$table";
+            }
+            else {
+                $class = ref $self;
+                $field = $value;
+            }
+
+            if ( $class->can('RecordClass')
+                and ( my $record_class = $class->RecordClass ) )
+            {
+                if ( my $meta = $record_class->_ClassAccessible->{$field} ) {
+                    if ( RT->Config->Get('DatabaseType') eq 'Pg' ) {
+                        if ( $meta->{is_numeric} || $meta->{type} eq 'datetime' ) {
+                            $value = "CAST($value AS VARCHAR)";
+                        }
+                    }
+                    elsif ( RT->Config->Get('DatabaseType') eq 'Oracle' ) {
+                        if ( $meta->{is_numeric} ) {
+                            $value = "TO_CHAR($value)";
+                        }
+                        elsif ( $type eq 'datetime' ) {
+                            $value = "TO_CHAR($value, 'YYYY-MM-DD HH24:MI:SS')";
+                        }
+                    }
+                }
+            }
+
+            if ( $type eq 'Date' ) {
+                $value = "SUBSTR($value, 1,  10)";
+            }
+        }
+        elsif ( $type eq 'IPAddress' ) {
             my $parsed = RT::ObjectCustomFieldValue->ParseIP($value);
             if ($parsed) {
                 $value = $parsed;
@@ -529,13 +656,6 @@ sub _LimitCustomField {
                 $RT::Logger->warn("$value is not a valid IPAddress");
             }
         } elsif ( $type eq 'IPAddressRange' ) {
-            if ( $value =~ /^\s*$RE{net}{CIDR}{IPv4}{-keep}\s*$/o ) {
-                # convert incomplete 192.168/24 to 192.168.0.0/24 format
-                $value =
-                  join( '.', map $_ || 0, ( split /\./, $1 )[ 0 .. 3 ] ) . "/$2"
-                  || $value;
-            }
-
             my ( $start_ip, $end_ip ) =
               RT::ObjectCustomFieldValue->ParseIPRange($value);
             if ( $start_ip && $end_ip ) {
@@ -604,7 +724,7 @@ sub _LimitCustomField {
         } elsif ( $type =~ /^Date(?:Time)?$/ ) {
             my $date = RT::Date->new( $self->CurrentUser );
             $date->Set( Format => 'unknown', Value => $value );
-            if ( $date->Unix ) {
+            if ( $date->IsSet ) {
                 if (
                        $type eq 'Date'
                            # Heuristics to determine if a date, and not
@@ -659,33 +779,10 @@ sub _LimitCustomField {
     }
 
     ########## Limits
-    # IS NULL and IS NOT NULL checks
-    if ( $op =~ /^IS( NOT)?$/i ) {
-        my ($ocfvalias, $CFs) = $self->_CustomFieldJoin( $cfkey, $cf, $ltype );
-        $self->_OpenParen( $args{SUBCLAUSE} );
-        $self->Limit(
-            %args,
-            ALIAS    => $ocfvalias,
-            FIELD    => ($column || 'id'),
-            OPERATOR => $op,
-            VALUE    => $value,
-        );
-        # See below for an explanation of this limit
-        $self->Limit(
-            ALIAS      => $CFs,
-            FIELD      => 'Name',
-            OPERATOR   => 'IS NOT',
-            VALUE      => 'NULL',
-            ENTRYAGGREGATOR => 'AND',
-            SUBCLAUSE  => $args{SUBCLAUSE},
-        ) if $CFs;
-        $self->_CloseParen( $args{SUBCLAUSE} );
-        return;
-    }
 
     my $single_value = !blessed($cf) || $cf->SingleValue;
     my $negative_op = ($op eq '!=' || $op =~ /\bNOT\b/i);
-    my $value_is_long = (length( Encode::encode_utf8($value)) > 255) ? 1 : 0;
+    my $value_is_long = (length( Encode::encode( "UTF-8", $value)) > 255) ? 1 : 0;
 
     $cfkey .= '.'. $self->{'_sql_multiple_cfs_index'}++
         if not $single_value and $op =~ /^(!?=|(NOT )?LIKE)$/i;
@@ -711,6 +808,7 @@ sub _LimitCustomField {
             OPERATOR   => $op,
             VALUE      => $value,
             CASESENSITIVE => 0,
+            QUOTEVALUE => $args{QUOTEVALUE},
         ) );
         $self->Limit(
             %args,
@@ -763,17 +861,18 @@ sub _LimitCustomField {
         );
     } else {
         # Otherwise, go looking at the Content
-        $self->Limit(
+        $self->Limit( $fix_op->(
             %args,
             ALIAS    => $ocfvalias,
             FIELD    => 'Content',
             OPERATOR => $op,
             VALUE    => $value,
             CASESENSITIVE => 0,
-        );
+        ) );
     }
 
-    if (!$value_is_long and $op eq "=") {
+    if ( ( blessed($cf) and $cf->IsNumeric ) or ( !$value_is_long and $op eq "=" ) ) {
+        # Skip LargeContent comparison for numeric values.
         # Doesn't matter what LargeContent contains, as it cannot match
         # the short value.
     } elsif (!$value_is_long and $op =~ /^(!=|<>)$/) {
@@ -815,6 +914,7 @@ sub _LimitCustomField {
             ENTRYAGGREGATOR => 'AND',
             SUBCLAUSE       => $args{SUBCLAUSE},
             CASESENSITIVE => 0,
+            QUOTEVALUE      => $args{QUOTEVALUE},
         ) );
         $self->_CloseParen( $args{SUBCLAUSE} ); # LargeContent check
     }
@@ -859,7 +959,7 @@ sub _LimitCustomField {
 
 This Limit sub calls SUPER::Limit, but defaults "CASESENSITIVE" to 1, thus
 making sure that by default lots of things don't do extra work trying to 
-match lower(colname) agaist lc($val);
+match lower(colname) against lc($val);
 
 We also force VALUE to C<NULL> when the OPERATOR is C<IS> or C<IS NOT>.
 This ensures that we don't pass invalid SQL to the database or allow SQL
@@ -875,10 +975,6 @@ my %check_case_sensitivity = (
 );
 
 my %deprecated = (
-    groups => {
-        type => 'Name',
-    },
-    principals => { objectid => 'id' },
 );
 
 sub Limit {
@@ -896,13 +992,14 @@ sub Limit {
     }
 
     if (($ARGS{FIELD}||'') =~ /\W/
-          or $ARGS{OPERATOR} !~ /^(=|<|>|!=|<>|<=|>=
+          or $ARGS{OPERATOR} !~ /^((?:SHALLOW\s*)?(?:=|<|>|!=|<>|<=|>=
                                   |(NOT\s*)?LIKE
                                   |(NOT\s*)?(STARTS|ENDS)WITH
                                   |(NOT\s*)?MATCHES
                                   |IS(\s*NOT)?
                                   |(NOT\s*)?IN
-                                  |\@\@)$/ix) {
+                                  |\@\@
+                                  |AGAINST))$/ix) {
         $RT::Logger->crit("Possible SQL injection attack: $ARGS{FIELD} $ARGS{OPERATOR}");
         %ARGS = (
             %ARGS,
@@ -921,7 +1018,7 @@ sub Limit {
     if ( $table and $ARGS{FIELD} and my $instead = $deprecated{ lc $table }{ lc $ARGS{'FIELD'} } ) {
         RT->Deprecated(
             Message => "$table.$ARGS{'FIELD'} column is deprecated",
-            Instead => $instead, Remove => '4.4'
+            Instead => $instead, Remove => '5.2'
         );
     }
 
@@ -937,6 +1034,16 @@ sub Limit {
         $ARGS{'CASESENSITIVE'} = 1;
     }
 
+    # Oracle doesn't support to directly compare CLOB with VARCHAR/INTEGER.
+    # DefaultDashboard search in RT::Dashboard::CurrentUserCanDelete needs this
+    if (   $ARGS{OPERATOR} !~ /IS/i
+        && !$ARGS{FUNCTION}
+        && RT->Config->Get('DatabaseType') eq 'Oracle'
+        && $self->RecordClass->_Accessible( $ARGS{FIELD}, 'is_blob' ) )
+    {
+        $ARGS{FUNCTION} = 'TO_CHAR(?)';
+    }
+
     return $self->SUPER::Limit( %ARGS );
 }
 
@@ -1051,6 +1158,39 @@ sub NotSetDateToNullFunction {
     return $res;
 }
 
+sub DistinctFieldValues {
+    my $self  = shift;
+    my %args = (
+        Field       => undef,
+        Order       => undef,
+        Max         => undef,
+        decode_utf8 => 1,
+        @_%2 ? (Field => @_) : (@_)
+    );
+
+    my @values = $self->SUPER::DistinctFieldValues( %args );
+
+    foreach my $value ( @values ) {
+        if ( $args{'decode_utf8'} ) {
+            if ( !utf8::is_utf8( $value ) ) { # mysql/sqlite
+                utf8::decode( $value );
+            }
+        }
+        else {
+            if ( utf8::is_utf8( $value ) ) {
+                utf8::encode( $value );
+            }
+        }
+    }
+    return @values;
+}
+
+sub CurrentUserCanSeeAll {
+    my $self = shift;
+    return 1 if $self->{_current_user_can_see_all};
+    return $self->CurrentUser->HasRight( Right => 'SuperUser', Object => RT->System ) ? 1 : 0;
+}
+
 RT::Base->_ImportOverlays();
 
 1;
